# Copyright(C)  2004 by Intevation GmbH
# Authors:
# Ludwig Reiter <ludwig@intevation.de>
# Silke Reimer <silke@intevation.de>
#
# This program is free software under the GPL(>=v2)
# Read the file COPYING coming with the software for details.
"""Hier sind die Klassen fr edbsilon.
   Reihenfolge der Hauptsachen:
               globale Variablen
               punkttrans
               Klasse edbs_line
               Klasse Linienzug
               Klasse Polygon
               Klasse Loch
               Klasse Feature
               Klasse edbs_satz
               Klasse edbs_satz_liste  """

#fr die Winkelberechnungen
import math

import logging
log = logging.getLogger('')

#fr die Umwandelung des EDBS-Zeichensatzes
translation = (
    """??????????\n??\n??"""
    """????????????????"""
    """ !"#$%&'()*+,-./"""
    """0123456789:;<=>?"""
    """@ABCDEFGHIJKLMNO"""
    """PQRSTUVWXYZ^_"""
    """`abcdefghijklmno"""
    """pqrstuvwxyz?"""
    """????????????????"""
    """????????????????"""
    """????????????????"""
    """????????????????"""
    """????????????????"""
    """????????????????"""
    """????????????????"""
    """????????????????"""
)


format = "Oracle" #welche DB genuzt werden soll
Spezi ={} # Spezifikation der Daten: Je nach Bundesland verschieden
Tabnamen =[] # Liste der Namen der Tabellen, die vorkommen 
Punktfehler = ""

def punkttrans(pktstring):
    """Transferiert einen ALK-Punktstring zu einen Paar von x,y-Koordinaten"""
    global Punktfehler 

    if  ' ' in pktstring:
        Punktfehler += "error: In Punkttrans tritt ein nicht Punktstring auf '"
        Punktfehler += pktstring
        Punktfehler += "'\n"
        x= 3400000
        y= 5700000    

    else:
        x = 0.0
        y = 0.0
        x = 100000.0 * int(pktstring[0:2])
        y = 100000.0 * int(pktstring[2:4])
        x +=  1000.0 * int(pktstring[4:6])
        y +=  1000.0 * int(pktstring[6:8])
        x +=     0.001 * int(pktstring[8:14])
        y +=     0.001 * int(pktstring[14:20])
    return(round(x,3),round(y,3))      
 

class edbs_line:
    """Beinhaltet eine Teillinie eines Objekt oder 
      Teilobjekt. Damit dann die ganze Geometrie eines Objektes zusammen 
      is, muessen spaeter diese Linienobjekte zusammengesetzt werden."""

    def __init__(self, liniendaten, Reihenfolge="N"):
        """bergeben wird ein Tupel der Daten wie bei setLiniendata und die
        Reihenfolge.  Die Reihenfolge gibt an ob die Reihenfolge der Punkte
        dieser Linie noch gendert werden muss.  Ist die Reihenfolge "R" ,
        dann wird die Reihenfolge in der die Pkte kommen gendert und dann
        setLiniendata aufgerufen.  Ist die Reihenfolge "N" wird nur
        setLiniendata aufgerufen und die Reihenfolge der Punkte nicht mehr
        gendert
        """

        if Reihenfolge == "R":
            liniendaten2 = []
            liniendaten2.append(liniendaten[1])
            liniendaten2.append(liniendaten[0])
            liniendaten2.append(liniendaten[2])
            Lagepara = liniendaten[3]
            Lagepara.reverse()
            liniendaten2.append(Lagepara)
            liniendaten2.append(liniendaten[4])
            liniendaten2.append(liniendaten[5])
            liniendaten2.append(liniendaten[6])
            self.setLiniendaten(liniendaten2)
        elif Reihenfolge == "N":
            self.setLiniendaten(liniendaten)
        else:
            raise "error: Unexcepted R in construktor of line"

    def setLiniendaten(self,liniendaten):
        """bergeben wird ein Liste mit folgenden Daten:
           *Anfangspkt als EDBSString
           *Endpkt als EDBSString
           *Art der Liniengeometrie
           *Lageparametreliste =[Pkt als EDBS_String]
           *Folie
           *Linienart
           *Linienteilung(Links)
         und das wird gespeichert"""
        self.Anfangspkt = liniendaten[0]
        self.Endpkt = liniendaten[1]
        self.ArtderLiniengeo = liniendaten[2]
        self.Lageparaliste = liniendaten[3]
        self.Folie =liniendaten[4]
        self.Art =liniendaten[5]
        self.Linieteilung = liniendaten[6] 

    def get_Anfangspkt(self):
        """gibt den Anfangspunkt zurck"""
        return self.Anfangspkt

    def get_Endpkt(self):
        """gibt den Endpkt zurck"""    
        return self.Endpkt

    def get_ArtderLiniengeometrie(self):
        """gibt die Art der Liniengeometrie zurck. 
         2Byte gibt an was in der Lageparameter fr Daten sind"""
        return self.ArtderLiniengeo

    def get_Lageparaliste(self):
        """gibt die Lageparameterliste zurck"""
        return self.Lageparaliste

    def get_Linienart(self):
        """gibt die 4Byte Linienart zurck, die nach OS kodiert ist"""
        return self.Art

    def get_Linienteilung(self):
        """gibt die 2Byte Linienteilung der Fkt der Linie zurck"""
        return self.Linieteilung

    def get_Vektor(self):
        """Macht aus den ersten zwei Punkten einen Linien-Vektor"""

        # Hole Anfangs- und Endpunkt, Endpunkt ist davon abhngig, ob ein 
        # Lageparameter vorhanden ist oder nicht
        anfPunkt = self.get_Anfangspkt()
        if self.get_Lageparaliste() == []:
            endPunkt = self.get_Endpkt()
        else:
            endPunkt = self.get_Lageparaliste()[0]

        # Transformation der EDBS-Schreibweise in X- und Y- Koordinate
        transEndPunkt = punkttrans(endPunkt)
        transAnfPunkt = punkttrans(anfPunkt)

        # Berechnung des Vektors
        vektor = (transEndPunkt[0] - transAnfPunkt[0],
                transEndPunkt[1] - transAnfPunkt[1])

        return vektor 

    def winkel(self,vektor, Linie, winkelart):
        """Bei winkelart == 'I': ist der Innenwinkel von self kleiner als
        der von Linie bezglich des Vergleichsvektors?  Bei winkelart ==
        'A': ist der Aussenwinkel von self kleiner als der von Linie bzg
        Vergleichsvektor.  Annahme: der Vergleichsvektor ist schon in
        Koordinaten transformiert.
           """

        # Existiert keine Vergleichslinie ist man fertig    
        if Linie == None:
            return 1

        # macht aus den letzten beiden Punkten der Linie self
        # einen Einheitsvektor    
        r_1 = self.get_Vektor()        
        r_1_Laenge = math.sqrt(math.pow(r_1[0],2) + math.pow(r_1[1],2))
        ein_r_1 =(r_1[0] /r_1_Laenge,r_1[1] / r_1_Laenge)

        # macht aus den letzten beiden Punkten der Linie Linie
        # einen Einheitsvektor
        r_2 = Linie.get_Vektor()
        r_2_Laenge = math.sqrt(math.pow(r_2[0],2) + math.pow(r_2[1],2))
        ein_r_2 =(r_2[0] /r_2_Laenge,r_2[1] / r_2_Laenge)
        
        # erzeugt die winkel zwischen Vergleichsvektor und den Einheitsvektoren
        winkelein_r_1 = vektor[0] * ein_r_1[0] + vektor[1] *ein_r_1[1]
        winkelein_r_2 = vektor[0] * ein_r_2[0] + vektor[1] *ein_r_2[1]
        
        # erzeugt den winkel zwischen um 90Grad verdrehten Vergleichsvektor
        # und den Einheitsvektoren
        winkelgedrehtein_r_1 = vektor[1] * ein_r_1[0] - vektor[0] * ein_r_1[1]
        winkelgedrehtein_r_2 = vektor[1] * ein_r_2[0] - vektor[0] * ein_r_2[1]
        
        # Je nach winkelart, die berprft werden soll, wird das Ergebnis
        # je Fall zurckgegeben
        if winkelart == "I":
            if winkelgedrehtein_r_1 >0 and winkelgedrehtein_r_2 <0:
                return 1
            elif winkelgedrehtein_r_1 >0 and winkelein_r_1 <= winkelein_r_2:
                return 1
            elif winkelgedrehtein_r_1 == 0 and winkelein_r_1 <= winkelein_r_2:
                return 1    
            elif winkelgedrehtein_r_1 <= 0 and winkelgedrehtein_r_2 <= 0 and winkelein_r_1 >= winkelein_r_2:       
                return 1
            else:
                return 0
        elif winkelart == "A":
            if winkelgedrehtein_r_1 <0 and winkelgedrehtein_r_2 >0:
                return 1
            elif winkelgedrehtein_r_1 <0 and winkelein_r_1 <= winkelein_r_2:
                return 1
            elif winkelgedrehtein_r_1 == 0 and winkelein_r_1 <= winkelein_r_2:
                return    
            elif winkelgedrehtein_r_1 >= 0 and winkelgedrehtein_r_2 >= 0 and winkelein_r_1 <= winkelein_r_2:       
                return 1
            else:
                return 0             

# Folgende Klassen sind fr den Feature 
class Linienzug:
    """Diese Klasse ist ein Kontainer fr eine Liste von Punkten, die
    eine Linie darstellen."""

    def __init__(self):
        """Initalisiert die Punktliste"""
        self.pktlist =[]

    def vorneeinfuegen(self , Punkt):
        """fgt Punkt  vorne in die Punktliste ein"""
        self.pktlist = [ Punkt ] +self.pktlist

    def hinteneinfuegen(self, Punkt):
        """fgt Punkt hinten in die Punktliste ein"""
        self.pktlist.append(Punkt)

    def getAnfangspkt(self):
        """gibt den ersten Punkt der Punktliste zurck"""
        return self.pktlist[0]

    def getEndpkt(self):
        """gibt den letzten Punkt der Punktliste zurck"""
        return self.pktlist[-1]

    def getganzeListe(self):
        """gibt die ganz Punktliste zurck"""
        return self.pktlist

class Ring:
    """Kontainer fr die Punkte eines Ringes"""
    def __init__(self, list = []):
        """Initialisiert die Punktliste"""
        self.pktlist = list
    def einfuegen(self,Punkt):
        """fgt einen Punkt  hinten in die pktlist an"""
        self.pktlist.append(Punkt)
    def getEndpkt(self):
        """gibt den letzten Punkt der pktlist zurck"""
        return self.pktlist[-1]
    def complete(self):
        """schaut ob Anfangs- und Endpkt zusammenpassen"""
        return self.pktlist[0] == self.pktlist[-1] 
    def getganzeListe(self):
        """Gibt die ganze Punktliste zurck"""
        return self.pktlist
    def isPolygon(self):

        area = 0
        pktList = self.getganzeListe()
        (refx, refy) = pktList[0]
        for i in range(len(pktList)):
            (p1x, p1y) = pktList[i]
            (p2x, p2y) = pktList[i-1]
            area += (p1x - refx) * (p2y - refy) - (p2x - refx) * (p1y - refy)
        return area > 0


class Polygon:
    """Kontainer fr ein Polygon mit Lchern"""
    def __init__(self, ring = []):
        """Initialisiert die Punkt- und die Lcherliste"""
        self.pktlist = ring
        self.loecher = []
    def einfuegen(self, Punkt):
        """Fgt einen Punkt hinten in der pktlist ein"""
        self.pktlist.append(Punkt)
    def einfuegenLoch(self, Loch):
        """Fgt ein fertiges Loch in der Lochliste ein"""
        self.loecher.append(Loch)
    def getEndpkt(self):
        """gibt den letzten Punkt der Pktliste zurck"""
        return self.pktlist[-1]
    def getganzeListe(self):
        """gibt die ganzePunktliste zurck"""
        return self.pktlist
    def getLoecher(self):
        """gibt die Lcher zurck"""
        return self.loecher

class Loch:
    """Kontainer fr die Punkte eines Loches"""
    def __init__(self, ring = []):
        """Initialisiert die Punktliste"""
        self.pktlist = ring
    def einfuegen(self,Punkt):
        """fgt einen Punkt  hinten in die pktlist an"""
        self.pktlist.append(Punkt)
    def getEndpkt(self):
        """gibt den letzten Punkt der pktlist zurck"""
        return self.pktlist[-1]
    def getganzeListe(self):
        """Gibt die ganze Punktliste zurck"""
        return self.pktlist

class Feature:
    """Ein Objektdaten-Sammler speicher alle Information zu einem
    EDBS-Objekt, d.h. alle Linieninformationen, alle Besondereren
    Informationen und alle Teilobjekte.

    Ein Objekt hat ein Status, der angibt ob beim Eintragen
    das Objekt in der DB vorher gelscht werden soll , nur gelscht
    werden soll, oder verndert werden soll oder neu eingetragen.
    Folgende Ausprgungen hat das Statusbyte: 
          N: Neueintrag
          L: Objekt wird gelscht
          M: Objekt wird gelscht und danach neu eingetragen
          E: Objekt wird gendert

    Die gamma-Daten sind in self.II_daten in der Form : 
        (FunktiondesObjektes,[([BesondereInfos], [Geometrieangaben])])
    gespeichert.
    """

    def __init__(self, Objnr):
       """Konstruktor: Das Statusfeld wird auf N gesetzt
        und die Objnr gespeichert."""
       self.Status ="N"
       self.Objnr = Objnr
       self.linelist = []
       self.Objkoor = (0.0 , 0.0)
       self.II_daten = None
       self.AK = None
       self.folie = ''
       self.art = ''
       self.aktualitaet = ''
       self.typ = ''
       self.datum = ''
       self.tag = ''
       self.monat = ''
       self.jahr = ''
       self.objektteile = []

    def compareObjnr(self , Objnr):
        """True, wenn 'Objnr' mit der Objektnr des Objektes identisch ist"""
        return self.Objnr == Objnr

    def appendline(self, line):
        """Fgt eine neue line-Objekt am Ende der Liste von Feature
        ein.  Falls das Statusfeld auf 'L', nderte das nach 'M' """
        self.linelist += [line]
        if self.Status == 'L':
           self.Status = 'M'
       

    def loescheObj(self,AK):
       """Setzt das Statusfeld auf 'L', lscht alle Daten in Feature"""

       if self.AK == None or self.AK == AK:
           self.Status = 'L'
           self.II_daten = None
           self.linelist = []
           self.AK = AK

    def setObjdaten(self, daten, Objpunkt):
       """Setzt die Objektdaten mit Hilfe eines gamma-Teils des Objekt"""
       self.II_daten  = daten
       self.Objkoor = Objpunkt
       self.folie = daten[0][0:3]
       self.art = daten[0][3:7]
       self.AK = daten[0][7:9]
       self.typ = daten[0][9:10]
       self.jahr = daten[0][17:19]
       self.monat = daten[0][19:21]
       self.tag = daten[0][21:23]
       self.datum = daten[0][17:23]
       self.objektteile = self.extractOT(daten[1])

       if self.Status =='L':
           self.Status = 'M'

    def extractOT(self, daten):

        otList = []
        for d in daten:
            ot = objektTeil()
            ot.setDaten(d)
            otList.append(ot)
        return otList

    def get_ganzeLinie(self):
        """Die einzelnen edbs_line-Objekte stellen, wenn das grosse Objekt
        eine Linie ist, nur Teilstcke dar. Diese Teilstcke werden von
        dieser Funktion in verschiedene Linienzge zusammengefasst.  Die
        Rckgabe ist eine Liste von Linienzug-Objekten"""

        # Hier werden die benutzten Datenstrukturen initialisiert
        # queue ist eine Warteschlange von edbs_line -Objekten
        # Diese Objekte sollen zu einer ganzen Linie zusammengesetzt werden.
        queue = self.linelist[:]
        # ergebnis beinhaltet die Linienzge, die aus den
        #einzelnen edbs_line-Linien zusammengesetzt sind
        ergebnis =[]
        count =0

        # Solange noch Linienteile zu verarbeiten sind, also der queue nicht
        # leer ist, gehe diese Ababeitungsschleife durch        
        while queue != []:

            # Das zu bearbeitende Linienstck wird aus den queue entnommen 
            linie = queue[0]
            anfpkt = punkttrans(linie.get_Anfangspkt())
            endpkt = punkttrans(linie.get_Endpkt())
            Art = linie.get_Linienart()
            del queue[0]

            # Eintrag der ersten Linie
            if ergebnis == []:
                # Am Anfang soll keine temporre Linie eingefgt werden
                if Art == '0242':
                    queue.append(linie)
                # der erste Lineinzug wird eingefgt
                else:
                    count =0
                    Zug = Linienzug() # Der erste Linienzug
                    Zug.hinteneinfuegen(anfpkt)
                    for pkt in linie.get_Lageparaliste():
                        Zug.hinteneinfuegen(punkttrans(pkt))
                    Zug.hinteneinfuegen(endpkt)
                    ergebnis = [ Zug ]

            # Linienstcke werden immer nur an die letzte Linie in der
            # Ergebnisliste eingefgt. Eine temporre Linie, die von einer
            # anderen Linie weggeht, beendet diese automatisch.
            else:
                ergebnisline = ergebnis[-1]
                # Endpunkt der letzten Linie und Anfangspkt der neuen
                # stimmen berein
                ergEndpkt = ergebnisline.getEndpkt()
                ergAnfpkt = ergebnisline.getAnfangspkt()
                if ergEndpkt == anfpkt or ergEndpkt == endpkt:
                    count = 0
                    # temporre Linie wird ignoriert, wenn sie zurck fhrt
                    # zur vorletzten Linie. Sonst fhrt sie zum Anlegen
                    # eines neuen Liniezuges.
                    if Art == "0242":
                        # Temporre Linie fhrt lediglich zurck zur
                        # letzten Linie
                        if len(ergebnis) >= 2 \
                            and (ergebnis[-2].getAnfangspkt() == anfpkt\
                            or ergebnis[-2].getAnfangspkt() == endpkt\
                            or ergebnis[-2].getEndpkt() == endpkt\
                            or ergebnis[-2].getEndpkt() == anfpkt):
                            pass
                        else:
                            Zug = Linienzug()
                            if ergEndpkt == anfpkt:
                                Zug.hinteneinfuegen(endpkt)
                            else:
                                Zug.hinteneinfuegen(anfpkt)
                            ergebnis += [ Zug ]
                    # anderen Linienstcke werden einfach an den letzten
                    # Liniezug angehngt
                    else:
                        tmp_liste = linie.get_Lageparaliste()
                        lastpoint = endpkt
                        if ergEndpkt == endpkt:
                            tmp_liste.reverse()
                            lastpoint = anfpkt
                        for pkt in tmp_liste:
                            ergebnis[-1].hinteneinfuegen(punkttrans(pkt))
                        ergebnis[-1].hinteneinfuegen(lastpoint)
                elif ergAnfpkt == endpkt or ergAnfpkt == anfpkt:
                    count = 0
                    # temporre Linie wird ignoriert, wenn sie zurck fhrt
                    # zur vorletzten Linie. Sonst fhrt sie zum Anlegen
                    # eines neuen Liniezuges.
                    if Art == "0242":
                        # Temporre Linie fhrt lediglich zurck zur
                        # letzten Linie
                        if len(ergebnis) >= 2 \
                            and (ergebnis[-2].getAnfangspkt() == anfpkt\
                            or ergebnis[-2].getAnfangspkt() == endpkt\
                            or ergebnis[-2].getEndpkt() == endpkt\
                            or ergebnis[-2].getEndpkt() == anfpkt):
                            pass
                        else:
                            Zug = Linienzug()
                            if ergAnfpkt == endpkt:
                                Zug.hinteneinfuegen(anfpkt)
                            else:
                                Zug.hinteneinfuegen(endpkt)
                            ergebnis += [ Zug ]
                    else:
                        # Punkte werden in umgekehrter Reihenfolge am
                        # vorderen Ende der Linie eingefgt
                        tmp_liste =linie.get_Lageparaliste()
                        firstpoint = anfpkt
                        if ergAnfpkt == anfpkt:
                            tmp_liste.reverse()
                            lastpoint = endpkt
                        for pkt in tmp_liste:
                            ergebnis[0].vorneeinfuegen(punkttrans(pkt))
                        ergebnis[-1].vorneeinfuegen(firstpoint)
                # Linie passt nicht, und wird wieder hinten an die queue
                # gehngt. Wenn alle Linienstcke bearbeitet wurden ohne zu
                # passen, wird abgebrochen.
                else:
                    if count < len(queue):
                        queue.append(linie)
                    else:
                        # TODO: wenn nur templinien brig sind, gibt es
                        # kein Problem
                        if queue != []:
                            log.warning("Warnung: Linienzug kann nicht vervollstndigt werden "+self.Objnr +"\n" )
                            count = 0
                            break
                    count +=1

        # Falls Anfangspkt des ersten Linienzugs und Endpkt des letzten
        # Linienzugs gleich sind, wird eine Warung ausgegeben
        if ergebnis[0].getAnfangspkt() == ergebnis[-1].getEndpkt():
            log.warning("Warnung: Der Anfangs- und Endpunkt der Linie fr Objekt %s sind identisch \n" % self.Objnr)

        return ergebnis

    def get_Polygon(self): 
        """Die einzelnen edbs_line-Objekte in einen Polygon stellen nur 
        Teilstcke dar. Diese Teilstcke werden hier zusammengefasst.
        Es kann auch vorkommen, dass ein Polygon Lcher hat. 
        Die Rckgabe ist eine Liste von Polygonen

        Der Algorithmus sollte folgendermassen sein:
        Man macht alles erst mit den Innenwinkel. Wenn das nicht klappt nochmal
        mit dem Aussenwinkel statt dem Innenwinkel:
        Zuerst baut man ein Polygon -Objekt auf und schreibt es in die
        Arbeitsliste.
        Man geht die noch nicht gearbeiteten Linienstcke durch und findet die 
        darzustellendes Liniestcke mit kleinsten Innenwinkel und
        die mglicherweise vorhandene nichtdarzustellende Liniestck, 
        die von den Endpkt losgeht.
        Jetzt gibt es drei Flle:
        1. Es gibt eine Nichtdarzustellendes -Linienstck  und beim letzten
        Durchlauf wurde keine Nichtdarzustellendes Linienstck bearbeitet.
        2. Es gibt ein darzustellendes Liniestck
        3. Es gibt keins von beiden: dann muss man zu den zweiten Grobdurchgang
          schalten(den mit Aussenwinkel),wenn man alle mal durchgegangen
          ist
        Der 1. Fall gliedert sich weiter, je nachdem auf der ringlist an
        0.er Stelle ein Loch oder ein Polygon ist.
        Ausserdem wird nun die nichtdarzustellende Linie bearbeitet und kann
        aus dem queue geloescht werden.   
        1.1. Ist auf der Arbeitsliste ein Loch:
        1.1.1 Ist das Loch fertig: ja. Dann wird das Loch in das oberste 
        Polygon geschrieben 
        1.1.2 Nein, dann wird ein neues Loch begonnen
        1.2 Ist auf der Arbeitsliste ein Polygon,
        dann wird gecheckt ob das Polygon fertig ist:
        1.2.1. Wenn ja, dann wird es in das ergebnis gepackt
        1.2.2. Wenn nein, wird je nach verhltnis des darzustellendes und des
        nichtdarzustellenden Linienstckes ein neues Loch bzw Polygon gebaut und
        auf die ringlist geschickt
        2. Fall: Es wird das Linienstck an das ersten Element der ringlist
        angebaut.   
        Falls weder
        """

        # Der Algorithmus mus zweimal durchlaufen werden, wenn es 
        # mit den Innenwinkeln bei ersten Mal nicht klappt.
        # Beim zweiten mal benutzt man Aussenwinkel
        for buch in ["I","A"]: 

            queue = self.linelist[:]
            ringlist = []
            lochlist = []
            ergebnis = []
            count = 0

            # Zum erkennen ob vorher eine Nichtdarzustellende Linie
            # verarbeitet wurde 
            nodarflag = 0 

            # Trage die erste darzustellenden Linie als Linie in das
            # Gesamtpolygon (ringlist) ein
            for linie in queue:
                if linie.get_Linienart() != "0242":

                    ring = Ring([])
                    # Anfangspunkt
                    ring.einfuegen(punkttrans(linie.get_Anfangspkt()))
                    # alle Lageparameter
                    for pkt in linie.get_Lageparaliste():
                        ring.einfuegen(punkttrans(pkt))
                    # Endpunkt
                    ring.einfuegen(punkttrans(linie.get_Endpkt()))

                    # Polygon 
                    ringlist += [ ring ]
                    queue.remove(linie)

                    # Erzeuge den ersten Vergleichsvektor
                    temp_koord = ringlist[0].getganzeListe()
                    vektor =(temp_koord[-1][0] - temp_koord[-2][0] ,
                                         temp_koord[-1][1] - temp_koord[-2][1])
                if len(ringlist) > 0:
                    break

            while queue != [] and count < len(queue):

                if ringlist == []:
                    break 

                # Durchsuche noch nichtverarbeitete Linien nach
                # - passenden nichtdarzustellenden Linien
                # - passenden darzustellenden Linien, dort diejenige mit
                #   dem kleinsten Innenwinkel
                # Darzustellende linie mit mimimalen Innenwinkel
                darlinie= None
                # Nichtdarzustellende Linie
                nodarlinie = None
                for linie in queue:
                    anfPkt = punkttrans(linie.get_Anfangspkt())
                    vEndPkt = ringlist[0].getEndpkt()
                    if anfPkt == vEndPkt:
                        if linie.get_Linienart() == "0242":
                            if nodarflag == 0:
                                nodarlinie = linie
                        elif linie.winkel(vektor, darlinie, buch):
                            darlinie = linie

                if nodarlinie != None:

                    endPkt = punkttrans(nodarlinie.get_Endpkt())
                    anfPkt = punkttrans(nodarlinie.get_Anfangspkt())
                    # nodarline springt zu altem Objekt zurck
                    if len(ringlist) > 1 \
                        and endPkt == ringlist[1].getEndpkt():
                        ring = ringlist[0]
                        if not ring.complete():
                            print "Fehler: Ring sollte vollstndig sein!"

                        # -> Polygon wird in Ergebnis eingetragen
                        elif ring.isPolygon():
                            polygon = Polygon(ring.getganzeListe())
                            for loch in lochlist:
                                polygon.einfuegenLoch(loch)
                            lochlist = []
                            ergebnis += [ polygon ]

                            del ringlist[0]
                        # -> Loch wird in aktuelles Polygon eingetragen
                        # TODO: berprfen ob nodarline nicht auch 
                        # zwei Lcher verbinden kann
                        else:
                            loch = Loch(ring.getganzeListe())
                            del ringlist[0]
                            lochlist.append(loch)

                    # nodarline springt nicht zurck
                    # -> ein neuer Ring wird begonnen
                    else:
                        neuerRing = Ring( [ endPkt ])
                        ringlist = [ neuerRing ] + ringlist
                        nodarflag = 1

                    # neuen Vergleichsvektor bestimmen
                    anfPkt = punkttrans(nodarlinie.get_Anfangspkt())
                    endPkt = punkttrans(nodarlinie.get_Endpkt())
                    lagepara = nodarlinie.get_Lageparaliste()
                    if lagepara == []:
                        vektor =(endPkt[0] - anfPkt[0], endPkt[1] - anfPkt[1])
                    else:
                        lastPkt = punkttrans(lagepara[-1])
                        vektor =(endPkt[0] - lastPkt[0], endPkt[1] - lastPkt[1])

                    count = 0
                    queue.remove(nodarlinie)

                elif darlinie != None:

                    # neuen Punkt eintragen
                    anfPkt = punkttrans(darlinie.get_Anfangspkt())
                    endPkt = punkttrans(darlinie.get_Endpkt())
                    lagepara = darlinie.get_Lageparaliste()
                    for pkt in lagepara:
                        ringlist[0].einfuegen(punkttrans(pkt))
                    ringlist[0].einfuegen(endPkt)

                    # neuen Vergleichsvektor bestimmen
                    if lagepara == []:
                        vektor =(endPkt[0] - anfPkt[0], endPkt[1] - anfPkt[1])
                    else:
                        lastPkt = punkttrans(lagepara[-1])
                        vektor =(endPkt[0] - lastPkt[0], endPkt[1] - lastPkt[1])

                    count = 0
                    queue.remove(darlinie)
                    nodarflag = 0
                else:
                    count += 1
                    if count > len(queue):
                        log.warning("warning: Es muss mindestens eine darlinie -existieren")
                        break

            # Wenn der Algorithmus mit kleinstem Innenwinkel erfolgreich
            # war (d.h. alle Linienelemente sind verarbeitet, wird
            # abgebrochen.
            if queue == []:
                for ring in ringlist:
                    if ring.isPolygon():
                        polygon = Polygon(ring.getganzeListe())
                        for loch in lochlist:
                            polygon.einfuegenLoch(loch)
                        ergebnis += [ polygon ]
                    del ring
                return ergebnis

        return ergebnis

class objektTeil:

    def __init__(self):

        self.artderInfo = ''
        self.kartentyp = ''
        self.darstellung = ''
        self.BI = ''
        self.artderGeo = ''
        self.geoangabenListe = []


    def setDaten(self, daten):
        
        self.artderInfo = daten[0][0][0:2]
        self.kartentyp = daten[0][0][2:4]
        self.darstellung = daten[0][0][6:10]
        self.BI = daten[0][0][10:43]
        self.artderGeo = daten[0][0][43:45]

        self.geoangabenListe = daten[1]

    def geoangaben(self):

        """Die Geometrie-angaben werden je nach art(Art der Geometrie aus
        den besonderen Infos) bearbeitet:
        11,15: Die Linienpkte werden rausgelesen:
                  1. Pfeilspitze 
                  2. Pfeilanfang 
                  3. Knickpkte
               und in richtige Reihenfolge und Format gebracht.  Ausserdem
               werden sie mit punkttrans in richtige Koordinaten
               transformiert und ausgegeben [ Pfeilspitze ,(x1 , x2),..,
               Pfeilanfang ]

        51,52: Ein Pkt wird rausgelesen
                  Mit punkttrans transformiert.  liefert (x, y)
        53: Ein Winkel und ein Pkt wird gelesen. der pkt transformiert und
            ausgegeben(Winkel ,(pktx, pkty))"""

        if self.artderGeo in ('11', '15'):
            Linie =[]
            Linie +=[punkttrans(self.geoangabenListe[0])]
            for n in range(2,len(self.geoangabenListe)):
                Linie +=[punkttrans(self.geoangabenListe[n])]
            Linie +=[punkttrans(self.geoangabenListe[1])]
            return Linie

        elif self.artderGeo in ('51', '52'):
            return punkttrans(self.geoangabenListe[0])

        elif self.artderGeo == '53':
            return [float(self.geoangabenListe[0][2:7])*9/1000, punkttrans(self.geoangabenListe[1])]
        else:
           raise "error : die Art der Geometrieangeben ist unbekannt"


class II_data:
    """II_data verarbeitet die komplizierten Daten von get_II_data.
       Dann soll es mglich sein, die daten gut auszulesen."""
    def __init__(self, liste):
        """Konstruktor: Liest die Liste list ein und speichert  
        Objektkoor , alphaliste und gammaliste ab"""
        self.Objektkoor = liste[0]
        self.alphalist = liste[1]
        self.gammalist = liste[2]

    def get_alphalen(self):
        """gibt die Lnge der alphalist aus"""
        return len(self.alphalist)

    def get_gammalen(self):
        """gibt die Lnge der gammaliste aus"""
        return len(self.gammalist)

    def get_betalen(self):
        """gibt die Lnge der betaliste aus"""
        return len(self.gammalist[1])   

    def get_OIoderL(self):
        """gibt zurck ob Objekt infos in den Daten sind(gammateil)
           oder ob Linienteile drin sind(alphateil)
           O fr Objektinfos
           L fr Linieteil
           X fr beides leer
           Y fr beides voll""" 
        if   self.get_alphalen()  == 0 and self.get_gammalen() == 0:
            return "X"
        elif self.get_alphalen() != 0 and self.get_gammalen() == 0:
            return "L"
        elif self.get_alphalen() == 0 and self.get_gammalen() != 0:
            return "O"
        else:
            return "Y"

    def get_Objektkoor(self):
        """gibt die Objektkoordianten zurck"""
        return self.Objektkoor  

    def get_EndpktderLinie(self, alpha):
        """gibt den alpha-ten Endpunkt der Linie zurck, falls vorhanden
            sonst '' """
        if self.get_OIoderL() == "L"or self.get_OIoderL()=="Y":
            if alpha < self.get_alphalen():
                return self.alphalist[alpha][0]
            else:
                return ""
        else:
            return ""

    def get_ArtderLiniengeo(self, alpha):
        """gint die alpha-tte Art der Liniengeo zurck, falls vorhanden
        sonst '' """
        if self.get_OIoderL() == "L"or self.get_OIoderL()=="Y":
            if alpha < self.get_alphalen():
                return self.alphalist[alpha][1]
            else:
                return ""
        else:
            return ""

    def get_FktderLinieliste(self, alpha):
        """gibt die alpha -te Funktion der Line Liste zurck, falls
        vorhanden, sonst []"""

        if self.get_OIoderL() == "L"or self.get_OIoderL()=="Y":
           if alpha < self.get_alphalen():
                return self.alphalist[alpha][2]
           else:
               return []
        else:
            return []

    def get_Lageparaliste(self, alpha):
        """gibt die alpha -te Lageparameter-Liste zurck, falls vorhanden,
        sonst []"""

        if self.get_OIoderL() == "L"or self.get_OIoderL()=="Y":
           if alpha < self.get_alphalen():
                return self.alphalist[alpha][3]
           else:
               return []
        else:
           return []

    def get_FktdesObj(self, gamma):
        """gibt die gamma-te Funktion des Objektes zurck, falls vorhanden,
        sonst '' """

        if self.get_OIoderL() == "O"or self.get_OIoderL()=="Y":
           if gamma< self.get_gammalen():
                return self.gammalist[gamma][0]
           else:
               return ""
        else:
            return ""

    def get_BesondereInfo(self, gamma, beta):
        """gibt fr die gamma-ten Eintrag die beta-ten Besonderen
        Informationen zurck, falls vorhanden, sonst '' """

        if self.get_OIoderL() == "O" or self.get_OIoderL()=="Y":
           if gamma < self.get_gammalen():
                return self.gammalist[gamma][1][beta][0]
           else:
               return []
        else:
            return []

    def get_Geometrieangabenliste(self, gamma, beta):
        """gibt fr den gamma-ten Eintrag die beta-te GeometrieInfoliste
        aus"""

        if self.get_OIoderL() == "O"or self.get_OIoderL()=="Y":
           if gamma < self.get_gammalen():
                return self.gammalist[gamma][1][beta][1]
           else:
               return []
        else:
            return []



class edbs_satz:
    """ein Kontainer fr eine EDBS-Daten-Zeile

       Diese Klasse wird fr first_parsing(file) genutzt.Sie speichert eine
       Zeile eines EDBS-Datensatzes als self.EDBS_line und einen Wert
       self.pos der dazuda ist die II- inforamtionen rauszulesen mit der
       Methode get_II_data()"""

    def __init__(self, data):
        """Konstruktor von entry: Man bergibt eine EDBS-Zeile als String.
           Wandelt diese Zeile gemss translation in richtigen Zeichencode
           Setzt die pos auf 0"""      
        self.EDBS_line = data.translate(translation)
        self.pos =0

    def __str__(self):
        """Wenn das Objekt zum String werden soll,
        gibt es den EDBS-Satz zurck, 
        der unter self.EDBS_line gespeichert ist."""
        return self.EDBS_line

    def relation(self):
        """gibt das Zugehrigkeitsbyte zurck"""
        return self.EDBS_line[22]

    def read_WHF(self):
        """holt sich 4 Bytes von der Position self.pos von self.EDBS_line,
        wandelt die in int um und gibt er zurck, erhht self.pos um 4  """
        tmp= int(self.EDBS_line[self.pos:self.pos+4])
        self.pos +=4
        return tmp

    def read_Point(self):
        """gibt 20 Bytes von Position self.pos zurck und erhht self.pos 
        um 20""" 
        tmp = self.EDBS_line[self.pos: self.pos + 20]
        self.pos +=20
        return tmp

    def read_Fktofline(self):
        """gibt 23 Bytes von Position self.pos zurck und erhht self.pos 
        um 23"""        
        tmp = self.EDBS_line[self.pos: self.pos + 23] 
        self.pos +=23
        return tmp        

    def read_postionparameter(self):
        """gibt 20 Bytes von Position self.pos zurck und erhht self.pos 
        um 20"""
        tmp = self.EDBS_line[self.pos: self.pos +20 ] 
        self.pos +=20
        return tmp      

    def read_specialinfo(self):
        """gibt 45 Bytes von Position self.pos zurck und erhht self.pos 
        um 45"""
        tmp = self.EDBS_line[self.pos: self.pos +45 ] 
        self.pos +=45
        return tmp      

    def read_geometrie(self):
        """gibt 20 Bytes von Position self.pos zurck und erhht self.pos 
        um 20"""
        tmp = self.EDBS_line[self.pos: self.pos +20 ] 
        self.pos +=20
        return tmp                      

    def read_Function(self):
        """gibt 23 Bytes von Position self.pos zurck und erhht self.pos 
        um 23"""
        tmp = self.EDBS_line[self.pos: self.pos +23 ] 
        self.pos +=23
        return tmp

    def get_II(self):
        """gibt self.EDBS_line ab 36 Byte zurck"""
        return self.EDBS_line[36:]      

    def get_II_data(self):
        """returns a list of IN(Infoname) datas:Objkoord,EndpktderLine
        FunktionderLine,Lagepara,FunktiondesObjektes,BesondersInfo,
        Geometrieangabe Dabei wird der II-Teil des Strings self.EDBS_line
        durchgegangen.  Folgendermassen ist die Ergebinsslist aufgebaut:
        (Grundrisskennzeichen,[(endpos,linetyp,[fktderLine], [Lagepara])],
        [(FktdesObjektes, [([besondereInfos] , [geometrieangeben])])])
           """

        # Hier beginnt der relevante II-Abschnitt des EDBS-Satzes
        self.pos =36
        #first WHF==0001 normally
        if self.read_WHF() != 1: 
            raise  "error: WHF 1 at the beginning excepted"

        #second WHF_A ==0001 normally
        if self.read_WHF() != 1: 
            raise  "error: 2nd WHF not 1"
        objektkoordination = self.read_Point()

        # Remember the "Prfzeichen"
        self.pos += 1 

        # WHFalpha
        whfalpha = self.read_WHF() 
        # WHF alpha times
        listalpha=[]
        for n in range(whfalpha):                    
            if self.read_WHF() !=1:  #WHF_Bi == 1 normally
                raise "error: WHF in alpha cycle not 1"
            endpoi = self.read_Point() # gets "Endpunkt der Linie"
            linetyp =self.EDBS_line[self.pos: self.pos +2] #Remember:    
            self.pos +=2                #  "Art  der Liniengeometrie"          
            whftheta = self.read_WHF() #WHF_theta
            fktlinelist =[]
            for n2 in range(whftheta):
                # if self.read_WHF() != 1: #WHFCij ==1
                #    raise  "error: WHF in theta cycle not 1: " +str(self.pos)
                fktline=self.read_Fktofline()
                #if self.read_WHF() != 1 : #WHFHij ==1
                #   raise "error: second WHF in theta cycle no 1"
                fktlinelist += [fktline ]  
            whfDi =self.read_WHF()
            paralist =[]
            for n3 in range(whfDi):
                para =self.read_Point() 
                paralist += [para]
            listalpha += [(endpoi,linetyp, fktlinelist, paralist)]               
        whfgamma= self.read_WHF()  # WHFgamma
        #WHF gamma time
        #if whfgamma != 1:
        #   raise "error: ALK -data just have one 'Funktion des Objekts'"
        gammalist =[]
        for n3 in range(whfgamma):
            if self.read_WHF() != 1 :
                raise "error: gamma cylce WHF not 1"
            functionofobj = self.read_Function()
            whfbeta = self.read_WHF()
            betalist=[]
            for n4 in range(whfbeta):
                whfFij = self.read_WHF()
                specialinfolist =[]
                for n in range(whfFij):
                    specialinfolist += [ self.read_specialinfo() ]
                whfGij = self.read_WHF()
                geometrielist =[]
                for n2 in range(whfGij):           
                    geometrielist += [ self.read_geometrie() ]
                betalist += [(specialinfolist ,geometrielist)]    
            gammalist += [(functionofobj,betalist)]    
        return (objektkoordination, listalpha ,gammalist)
        
        
    def get_entryart(self):
        """Je nach OP und IN gibt er folgendes zurck:
        OP    | IN       |Rckgabe
        BSPE  |ULOBNN..  |OBJEKTDATEN
        FEIN  |ULOBNN..  |OBJEKTDATEN
        BINF  |ULOTEX..  |AUFTRAGSKENNDATEN
        OTEX  |ILOTEX..  |AUFTRAGSTEXTDATEN
        FGEO  |ULODATUM  |AENDERUNGSDATUM  (fr BZSN)
        FLOE  |DLOB2005  |OBJDATENLOESCHEN (fr BZSN)
        BKRT  |IBENKRT.  |GEBIETSKENNZEICHNUNG
        AKND  |ULQA0000  |AUFTRAGSKENNSATZ
        AEND  |........  |AUFTRAGSENDE
        sonst            |UNDEFINIERT"""

        OP = self.EDBS_line[12:16]
        IN = self.EDBS_line[28:36]
        if   OP == "BSPE"  and IN=="ULOBNN  ":
            return "OBJEKTDATEN"
        elif(OP == "BSPE" or OP == "FEIN") and IN == "ULO8ALK ": 
            return "OBJEKTDATEN"  
        elif OP == "BINF" and IN == "ULOTEX  ":
            return "AUFTRAGSKENNDATEN"
        elif OP == "OTEX" and IN == "ILOTEX  ":
            return "AUFTRAGSTEXTDATEN"
        elif OP == "FGEO" and IN == "ULODATUM":
            return "AENDERUNGSDATUM"
        elif OP == "FLOE" and IN == "DLOB2005":
            return "OBJDATENLOESCHEN"    
        elif OP == "BKRT" and IN == "IBENKRT ":
            return "GEBIETSKENNZEICHNUNG"
        elif OP == "AKND" and IN == "ULQA0000":
            return "AUFTRAGSKENNSATZ"
        elif OP == "AEND" and IN == "        ":
            return "AUFTRAGSENDE"
        else:
            return "UNDEFINIERT"

    def get_SK_Objnr(self):
        """parst das SK(Suchkriterium) des entry -Objektes und gibt eine
        Objnr und eine Aktualitten zurck
        """

        if self.EDBS_line[36:48] == 'ASDLOB2005EQ' \
                and self.EDBS_line[55:67] == 'ANDLOB2003EQ':
            return (self.EDBS_line[48:55], self.EDBS_line[67:69])
        elif self.EDBS_line[36:48] == 'ASDLOB2003EQ' \
                and self.EDBS_line[50:62] == 'ANDLOB2005EQ':
            return (self.EDBS_line[62:69], self.EDBS_line[48:50])
        else:
           print(self.EDBS_line[48:53], self.EDBS_line[63:65])
           raise "error: Nicht-Verstandene SK-Zeile"    


class edbs_satz_liste   :
     """container for edbs_satz objects which belong together"""

     def __init__(self,data):
         """constructor of edbs_satz_liste, needs an entry"""
         self.list = [data]

     def additionalentry(self, data):
         """appends ein entry Objekt an die Liste self.list"""
         self.list += [data]

     def get_entryart(self):
         """geht die Liste von entry self.list noch OP und IN durch.
         schaut ob die alle gleich sind, sonst gibt er UNDEFINIERT zurck"""

         art=self.list[0].get_entryart()
         for en in self.list:
             if en.get_entryart() != art:
                 print "edbs_satz_art: "+ art +" self.list_art: "+"en.getentryart()"+en.get_entryart() 
                 return "UNDEFINIERT"
         return art      

     def get_II_data(self):
         """nimmt alle entry in self.list und verbindet ihr II Teil zu einen
         grossen II-Teil der in ein neuen entry verarbeitet werden kann"""

         tmp_zusammenII =""  
         for ein in self.list:
            Laenge = int(str(ein)[4:8])
            tmp_zusammenII += ein.get_II()[0:(Laenge -24)]
            # Dies scheint zu klappen. Kann aber Fehler machen:             
            # tmp_zusammenII += ein.get_II().rstrip()               
         self.zusammenIIentry = edbs_satz(str(self.list[0])[0:36]+tmp_zusammenII)
         return self.zusammenIIentry.get_II_data()

     def __str__(self):
         """verbindet die str-Ausgabe von den verschiedenen entry Objekten
            aus self.list und gibt dies Verbundene zurck"""

         output =""
         for x in self.list:
             output += str(x)
         return output
