# Copyright (C)  2004 by Intevation GmbH
# Authors:
# Ludwig Reiter <ludwig@intevation.de>
# Silke Reimer <silke@intevation.de>
#
# This program is free software under the GPL (>=v2)
# Read the file COPYING coming with the software for details.

###################
#
# TODO Liste fr das gesamte Projekt
#
#   * Abschluss OGR-Export -> Tests schreiben
#   * Wahlmglichkeit des Interfaces im Hauptmen
#   * Export als Methoden an die einzelnen Feature und Objektteile
#      implementieren
#   * nderung der Tabellenspezifikation: 
#       redunte Haltung von Attributinformationen z.B. durch Aufsplitten
#       des Flurstckkennzeichens ermglichen
#   * frei whlbare Tabellennamen fr Feature-Tabellen ermglichen
#   * berarbeitung der Systems fr Fehlerberichte
#       -> benutze 'raise' statt 'logging'
#
##################

import sys
import os
sys.path +=[os.path.join(os.pardir, "Lib")]

import logging

import edbsclasses
import parse
import exportDB
"""
edbsilon.py -- Konvertierungsprogramm fr EDBS-Daten Befehl:

  python edbsilon [ -o <Ziel der SQL_Kom>] [-l <Ziel des Logbuchs> ]
            [-k <Konfigurationsdatei>] Dateien....   

Die Dateien sind die EDBS-Quelldateien, die abgearbeitet werden sollen.

Vorgehensweise beim Konvertiern: 
   1. Parsing: Suche nach EDBS-Datenstzen, die Zusammen gehren.
      Dies wird durch das A F E in den Zugehrigkeitsbyte angegeben. 
      Jeder EDBS-Satz wird in ein EDBS_satz -Objekt kopiert.
      Stze, die zusammen gehren, werden in ein EDBS_satz_liste -Objekt
      geschoben. Dabei bleibt die Reihenfolge erhalten.
      In EDBS_satz_liste werden diese EDBS_satz-Objekte in der Liste
      list gespeichter.
      Die EDBS_satz -Objekte und die
      EDBS_satz_liste -Objekte werden in der Liste plist
      eingehngt.

      z.B [E,  EL,E,...,E] 0 plist
           ^    ^
           |    |    E means an objekt of EDBS_satz
           |    two or more
           |     EDBS-data-lines, which belong together, are stored in 
           |     edbs_satz_liste
           a standalone EDBS-data-line 

   2. Parsing: Geht die Liste aus edbs_satz und edbs_satz_liste -Objekte durch:
               Falls OP und IN folgendes sind, mache:
               BSPE ULOBNN.. : Grundrissdatenverarbeiten
               FEIN ULOBNN.. : Grundrissdatenverarbeiten 
               BINF ULOTEX.. : Text ins Logbuch schreiben
               OTEX ULOTEX.. : "
               BKRT IBENKRT. : Gebietskennzeichen ins Logbuchschreiben
               AKND ULQA0000 : Auftragskennsatz ins Logbuch schreiben
               AEND ........ : Ende ins Logbuch schreiben   
   3. Wegschreiben: Das Die ONS wird durchgegangen und bei den enthalten 
        Feature -Objekte wird die Objektnummer kontrolliert mit compareObjnr
        und dann wird durch wandel() die Informationen in SQL verpackt."""

class Folie:

    def __init__(self, folienNr):

        self.folienNr = folienNr
        self.tabname = ''
        self.attlist = {}



def lese_Konfdatei(Konf):
    """Aus der Datei Konf werden die Daten gezogen um das Directory Spezi
         fr die Spezifikationen aufzubauen.
       Geht folgendermassen vor: Holt sich eine Zeile,
       Splitted die Kommas raus, verknpft die ersten drei zum Schlssel
       und das vierte kommt in Wert  dann wird das sechste
       nochmal bzg ; gesplitted und die diese Attributsliste an Wert appended.
       Dies sind alle Infos fr einen Eintrag in dem Directory Spezi:
       Schlssel und [Tabelle, Attributsliste]. 
       Ausserdem wir eine Liste aller Tabellen angelegt: edbsclasses.Tabnamen"""

    # Einlesen der Konfigurationsdatei
    try:
        konffile = open(Konf,"r")
    except:
        print "Konfigurationsdatei konnte nicht geffnet werden."
        hilfeausgeben()
        sys.exit()

    konflines = konffile.readlines()
    konffile.close()

    edbsclasses.Tabnamen =[]
    edbsclasses.tables = {}

    for line in konflines:
        # Ignorieren der Zeile, da Kommentarzeile
        if line[0] =="#":
            continue

        # Erstellen der Konf_satz_liste durch splitten
        # Dies ist die List einer Konfigurationsdateizeile     
        konflist =  line.split(",")

        # Ignorieren der Zeile, da in dieser Zeile steht, ob
        # Folientabellen angelegt werden mssen
        # Dies braucht nur auto_create
        if konflist[1] =="" and konflist[2] =="":
            folie = Folie(konflist[0])
            # Hier wird der Name der Folientabellen gespeichert
            if 'l' in konflist[4]: 
                #edbsclasses.Tabnamen += ["f"+konflist[0]+"_l"]
                edbsclasses.Tabnamen += [konflist[3]+"_l"]
            if 'p' in konflist[4]:
                #edbsclasses.Tabnamen += ["f"+konflist[0]+"_p"] 
                edbsclasses.Tabnamen += [konflist[3]+"_p"] 
            if 'f' in konflist[4]:
                #edbsclasses.Tabnamen += ["f"+konflist[0]+"_f"]
                edbsclasses.Tabnamen += [konflist[3]+"_f"]
            folie.tabname = konflist[3]
            edbsclasses.tables[konflist[0]] = folie
            continue  # Ignorieren der Zeile, da Daten frs Tabellenanlegn

        # Sind richtig viele Teile in der Zeile?
        if len(konflist) != 6:
            log.warning("warning: Falsche Anzahl an Argumenten in\
                        Konfigurationszeile: %s" % konflist)
            continue
        else:
            # Erzeugen der Spezi -Daten
            Schluessel = konflist[0] + konflist[1] + konflist[2]
            Wert = [konflist[3] ]
            Wert_tmp = konflist[5].split(';')
            for n in range(len(Wert_tmp) ):
                if n % 2 == 1:
                    Wert_tmp[n] = int(Wert_tmp[n])    
            Wert.append(Wert_tmp ) 
            edbsclasses.Spezi[Schluessel] = Wert

            # Anlegen des Tabellennamens in Tabnamen
            if konflist[3] != "":
                if "n" in konflist[4] \
                and not konflist[3] in edbsclasses.Tabnamen:
                    edbsclasses.Tabnamen += [ konflist[3] ]
                if "l" in konflist[4] \
                and not konflist[3]+"_l" in edbsclasses.Tabnamen:
                    edbsclasses.Tabnamen += [ konflist[3] + "_l" ]
                if "p" in konflist[4] \
                and not konflist[3] +"_p" in edbsclasses.Tabnamen:
                    edbsclasses.Tabnamen += [ konflist[3] + "_p" ]

def hilfeausgeben():
    """mit print wird die Hilfe ausgegeben"""

    print "edbsilon -- Konvertiert EDBS-Daten"
    print "Benutzung: python edbsilon.py [Optionen] Quelldatein..."
    print "Optionen: -o <Ziel>         -- SQL-Kommandos [Default: edbsilon.sql]"
    print "          -k <Konffile>     -- Konfigurationsdatei [Default: alk.konf]"
    print "          -l <Ziel_Logbuch> -- Logbuch [Default: edbsilon.log]" 
    print "          -h                -- fr diese Hilfe"


# Diese Befehle laufen beim Hauptaufruf.    
if __name__ == "__main__":

    verbraucht = 0 

    ziel = 'edbsilon.sql'
    logbuch = 'edbsilon.log'
    Konfdatei = 'alk.konf'
    Quell_Datei_Liste = []
    # Die Argumente des aufrufes verstehen 
    for n in range(len (sys.argv)):
        arg = sys.argv[n]
        if n == 0:
            continue
        elif verbraucht == 1:
            verbraucht = 0
            continue
        elif arg[0:2] == "-h":
            hilfeausgeben()
            sys.exit()     
        elif arg[0:2] == "-o":
            verbraucht = 1
            if n+1 < len (sys.argv): 
                ziel = sys.argv[n+1]
            else:
                hilfeausgeben()
                sys.exit(1)    
        elif arg[0:2] == "-l":
            verbraucht = 1
            if n+1 < len (sys.argv):
                logbuch = sys.argv[n+1]
            else:
                hilfeausgeben()
                sys.exit(1)  
        elif arg[0:2] == "-k":
            verbraucht = 1
            if n+1 < len (sys.argv):
                Konfdatei = sys.argv[n+1]
            else:
                hilfeausgeben()
                sys.exit(1)
        else:
            Quell_Datei_Liste += [ arg ]
    
    # Ist keine Quelle angegeben: Hilfe ausgeben
    if Quell_Datei_Liste == []:
        hilfeausgeben() 
        sys.exit()
    
    log = logging.getLogger('')
    try:
        hdlr = logging.FileHandler(logbuch)
    except:
        print "Logbuch konnte nicht geffnet werden."
        hilfeausgeben()
        sys.exit()
    
    log_format = logging.Formatter('%(message)s')
    hdlr.setFormatter(log_format)
    log.addHandler(hdlr) 
    log.setLevel(logging.INFO)
    
    # Einlesen der Konfigurationsdatei
    print "Lese Konfigurationsdatei"
    lese_Konfdatei(Konfdatei)   
    
    
    if edbsclasses.format in ('Oracle', 'PG'):
        sqlfile = open(ziel,'w')

    # Parsen der EDBS-Daten aus Quelle
    for quelle in Quell_Datei_Liste:
    
        print "Parse", quelle
        ONS = {}
        parse.parsen(ONS, quelle)
    
        # Ausgabe, hier als SQL-Kommandos
        if edbsclasses.format in ('Oracle', 'PG'):
            exportDB.exportiere_SQL(ONS , sqlfile)
        if edbsclasses.format == 'OGR':
            ogrCollection = exportOGR.OGRCollection(edbsclasses.Tabnamen)
            ogrCollection.readONS(ONS)
            ogrCollection.export()


    if edbsclasses.format in ('Oracle', 'PG'):
        # Baut die Indizes auf den Tabellen neu auf
        exportDB.anpassen_Index(edbsclasses.Tabnamen, sqlfile)  
    
