# Copyright (C)  2004 by Intevation GmbH
# Authors:
# Ludwig Reiter <ludwig@intevation.de>
#
# This program is free software under the GPL (>=v2)
# Read the file COPYING coming with the software for details.

"""Erzeugt automatisch CREATE-SQL-Kommandos aus dem Konfigurationsfile.
   Bentigt 2 Argumente : Quell-Konffile Ziel-SQLfile
   Beinhaltet Klasse Tab fr Tabellen
           und Liste tablist fr die verschiedenen Tabellenobjekte
   
   Liest eine Zeile des Konffiles und bearbeitet sie:
        Wenn Kommentar:
             nchste Zeile
        Wenn Folienspezifikation(d.h. nur Folie und Tabellenflags gesetzt):
             Wenn Folientabellen alle vorhanden
                nchste Zeile
             Sonst lege zustzliche Tabellen an
               (d.h. Tab anlegen und in tablist appenden)
        Wenn BesondereInfospezifikation ohne Tabellenname:
             Gehe die Folientabellen in tabliste durch und lege die
               Attribute an, falls sie noch nicht vorhanden
        Wenn BesondereInfospezifikation mit Tabellenname:
             Wenn Tabelle noch nicht vorhanden:
                  Lege Tabelle an und appende sie in tablist
             Wenn Attribute noch nicht vorhanden:
                  Lege Attribute in der Tabelle an            
    Nach diesen Durchlauf mssten alle anzulegenden Tabellen mit ihren 
    Attributen in tablist als Tab-Objekte stehen
    Gehe tablist durch und erzeuge die SQL_Kommandos
    
    Folientabellen haben die Namen f<foliennr><_p oder _l oder _f>
    und mindestens die Attribute objnr, objart, aktualitaet, edatum, geom.
    
    anderen Tabellen haben mindestens die Attribute objnr, artderinfo,
    kartentyp, darstellungsart.
    """           
             
import sys    

class Tab:
    """Kontainer fr den Tabellennamen und die Attribute + Art der Attribute
    einer Tabelle. Speichert dies solange bis alle Konfigurationsdateizeilen
    verarbeitet sind. Dann wird mit SQL_create die SQL -Create-Kommandos 
    erzeugt."""
    def __init__(self, Name, ArtderTabelle):
        """Name ist der Name der Tabelle
        ArtderTabelle gibt an was fr eine Tabelle erzeugt werden soll:
           F fr Folientabelle  mit Linien- oder Flchengeometrie
           FP fr Folientablle mit Punktgeometrie
           Z fr Zusatzdatentabelle ohne Geometrie
           G fr Zusatzdatentabelle mit Geometrie
           GP fr Zusatzdatentabelle mit Punktgeometrie"""
        self.Name = Name
        self.Attribute = []
        self.Attribute.append(("objnr", "VARCHAR(7)"))
        self.Attribute.append(("ogr_fid", "INTEGER"))
        if  ArtderTabelle =="F" or ArtderTabelle == "FP" :
            self.Attribute.append( ("objart", "INTEGER" ) )
            self.Attribute.append( ("aktualitaet", "VARCHAR(2)"))
            self.Attribute.append( ("edatum", "DATE"))
            self.Attribute.append( ("geom", "MDSYS.SDO_GEOMETRY") )
            if ArtderTabelle == "FP" :
                self.Attribute.append( ("winkel", "INTEGER"))      
        elif ArtderTabelle == 'Z' or ArtderTabelle =='G'or ArtderTabelle=="GP":
            self.Attribute.append( ("artderinfo", "INTEGER") )
            self.Attribute.append( ("kartentyp", "VARCHAR(2)" ))
            self.Attribute.append( ("darstellungsart", "INTEGER" ))
            if ArtderTabelle == "G":
                self.Attribute.append( ("geom", "MDSYS.SDO_GEOMETRY") )
            elif ArtderTabelle == "GP":
                self.Attribute.append( ("geom", "MDSYS.SDO_GEOMETRY") )
                self.Attribute.append( ("winkel", "INTEGER") )    
                
    def isin_Attribute(self, Attribut):
        """sucht self.Attribute nach den Attribut durch:
           1 falls vorhanden
           0 sonst"""
        for (Attname,dummy) in self.Attribute:
            if Attname == Attribut:
                return 1
        return 0                         
        
    def append_Attribut(self, Atttupel):
        """fgt ein Attributtupel der Form ( Attributsnamen, Attributsart) in
        self.Attribute ein, wenn noch nicht vorhanden."""
        if not self.isin_Attribute( Atttupel[0]) and Atttupel[0] != "":
            self.Attribute.append( Atttupel )     
        
    def get_Name(self):
        """Gibt self.Name zurck"""
        return self.Name        
        
    def create_SQL(self):
        """erzeugt das CREATE-SQL-Kommando und gibt es zurck"""
        SQL_Kom ="CREATE TABLE "
        SQL_Kom += self.Name
        SQL_Kom += "\n ( "
        Komma =0
        for (Attname, Attart) in  self.Attribute: #Attname fr Attributname
            if Komma ==1:                         #Attart fr Attributsart
                SQL_Kom += ","
            SQL_Kom += "\n     "
            SQL_Kom += Attname
            SQL_Kom += "     "
            SQL_Kom += Attart
            Komma =1
        SQL_Kom += "\n ); \n"                     
        return SQL_Kom
    
    def erzeuge_Metadata_SQL(self):
        """erzeugt die in die METADATA -Insert-SQL-Befehle und den
        Index fr diese Tabelle.
        Dies muss man aendern, falls man die Indizes anpassen will."""
        SQL_Kom = ""
        if self.Name[-2:] in ['_p','_l','_f' ]:
            # Anlegen der Metadaten 
            SQL_Kom = "INSERT INTO USER_SDO_GEOM_METADATA VALUES (\n '"
            SQL_Kom += self.Name
            SQL_Kom += "',\n'geom',\n"
            SQL_Kom += "MDSYS.SDO_DIM_ARRAY( \n"

            SQL_Kom += "   MDSYS.SDO_DIM_ELEMENT( 'X', 3345000, 3440000, 0.001 ), \n"
            SQL_Kom += "   MDSYS.SDO_DIM_ELEMENT( 'Y', 5682322, 5754034, 0.001 ) \n"
            SQL_Kom += "), NULL ); \n"
            # Erzeugen des Indexes
            SQL_Kom += "CREATE INDEX "
            SQL_Kom += self.Name
            SQL_Kom += "_idx"
            SQL_Kom += " ON "
            SQL_Kom += self.Name
            SQL_Kom += "(geom) INDEXTYPE IS MDSYS.SPATIAL_INDEX; \n"
        return SQL_Kom           

    def erzeuge_Sequence_SQL(self):
        """Erzeugt eine Sequence, der die ID einer
        Tabelle automatisch weiterzhlt"""

        seq_sql = """
CREATE SEQUENCE %s_seq
MINVALUE 1
MAXVALUE 999999999999999999999999999
START WITH 1
INCREMENT BY 1
CACHE 20;
     """

        return seq_sql %self.Name

    def erzeuge_Trigger_SQL(self):
        """Erzeugt eine Trigger, der die ID einer
        Tabelle automatisch weiterzhlt"""

        trigger_sql = """
create trigger %(name)s_trigger
before insert on %(name)s
for each row
begin
    select %(name)s_seq.nextval into :new.ogr_fid from dual;
end;
/
        """

        return trigger_sql %{'name':self.Name}

def name_in_tablist( Name):
    """Schaut ob Name in einer Tabelle Tab in tablist der Name ist.
    Gibt dann Tab zurck"""
    for tab in tablist:
        if Name == tab.get_Name():
            return tab
    return 0                 
        
def eintragen_Attribute( Name, Komma_split_5, Art):
    """Trgt die Attribute in die Tabellen ein und legt die Tabellen an, falls
    noch nicht vorhanden"""   
    if not name_in_tablist(Name):
        Tabobj =Tab( Name , Art)
        tablist.append( Tabobj )
    else:
         Tabobj = name_in_tablist( Name)        
    Attributlist = Komma_split_5.split(";")
    for n in range( len( Attributlist )/2):
        #print Name, Komma_split
        if Attributlist[2*n+1] != "INTEGER":
            Tabobj.append_Attribut( (Attributlist[2*n],
                                 "VARCHAR(" + Attributlist[2*n +1].strip()+")"))               
        else:
            Tabobj.append_Attribut( (Attributlist[2*n],
                                 "INTEGER"))
        
# Initialisierung der tablist.        
tablist =[]
try: 
    quelle = open( sys.argv[1] , "r" )
    Konf_satz_liste = quelle.readlines()
finally:
    quelle.close()
for Konf_satz in Konf_satz_liste:
    if Konf_satz[0] =='#':
        continue
    Komma_split =  Konf_satz.split(",")    
    if len(Komma_split) != 6:
        print "warning: Die  Konfigurationsdatei knnte falsch sein" + str(Komma_split)
        continue    
    if Komma_split[1] =="" and Komma_split[2] =="": 
        Name = "f" + Komma_split[0]
        if "n" in Komma_split[4] and not name_in_tablist(Name):
            tablist.append( Tab( Name ,"F"))
        if "p" in Komma_split[4] and not name_in_tablist(Name+"_p"):
            tablist.append( Tab( Name+"_p" ,"FP")) 
        if "l" in Komma_split[4] and not name_in_tablist(Name +"_l"):
            tablist.append( Tab( Name+"_l" ,"F"))
        if "f" in Komma_split[4] and not name_in_tablist(Name +"_f"):
            tablist.append( Tab( Name+"_f" ,"F"))              
    elif Komma_split[3] == "":
        Name = "f" + Komma_split[0]
       # print Name
        if "n" in Komma_split[4]:
            eintragen_Attribute( Name , "kartentyp;2" ,"F")
            eintragen_Attribute( Name , "darstellungsart;INTEGER" ,"F")
            eintragen_Attribute( Name ,Komma_split[5], "F")
        if "p" in Komma_split[4]:
            eintragen_Attribute( Name + "_p" , "kartentyp;2" ,"FP")
            eintragen_Attribute( Name + "_p", "darstellungsart;INTEGER" ,"FP")
            eintragen_Attribute( Name + "_p" ,Komma_split[5],"FP")
        if "l" in Komma_split[4]:
            eintragen_Attribute( Name + "_l" , "kartentyp;2" ,"F")
            eintragen_Attribute( Name + "_l", "darstellungsart;INTEGER" ,"F")
            eintragen_Attribute( Name + "_l" ,Komma_split[5],"F")
        if "f" in Komma_split[4]:
            eintragen_Attribute( Name + "_f" , "kartentyp;2" ,"F")
            eintragen_Attribute( Name + "_f", "darstellungsart;INTEGER" ,"F")
            eintragen_Attribute( Name + "_f" ,Komma_split[5],"F")   
    else:
        Name =  Komma_split[3]
        if "n" in Komma_split[4]:
            eintragen_Attribute( Name ,Komma_split[5], "Z")
        if "p" in Komma_split[4]:
            eintragen_Attribute( Name+"_p" ,Komma_split[5],"GP")
        if "l" in Komma_split[4]:
            eintragen_Attribute( Name+"_l" ,Komma_split[5],"G")
        if "f" in Komma_split[4]:
            eintragen_Attribute( Name+"_f" ,Komma_split[5],"G")         
                  
SQL_Kom = ""
# Erzeugen der CREATE -Kommandos                  
for tab in tablist:
    SQL_Kom += tab.create_SQL() 
# Erzeugen der Metadaten und Index -Komandos
for tab in tablist:
    SQL_Kom += tab.erzeuge_Metadata_SQL()       
    SQL_Kom += tab.erzeuge_Sequence_SQL()
    SQL_Kom += tab.erzeuge_Trigger_SQL()
try:
    ziel = open( sys.argv[2] , "w" ) 
    ziel.write(SQL_Kom)       
    ziel.close()
except:
    print SQL_Kom


        
